/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  |           =====>>>>> 2002 <<<<=====             |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (input.C)
  
  Written by: Damian J. J. Farnell

  Date:       Wed Jul 17 12:03:05 BST 2002

  Purpose:    Inputs the CCM script data

  Changes:
      - cout replaced by printf, little improvements (Joerg Sep2005)

  ===================================================

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "../include/Constants.h"
#include "../include/Externals.h"
#include "../include/FunctionPrototypes.h"
#include "../include/nr.h"
#include "../include/nrutil.h"

#if (PARALLEL_FLAG>0)
#include<mpi.h>
#endif

FILE *script;

void convert_numeral_to_char(int number, char *character);

// function prototypes.

void  moveAlongHome(void);
double readInFloat(void);
int   readInInteger(void);
void  setArray(int **typeArray);
void  definePointsInBT(void);

void input(int argc, char *argv[]) 
{
  
  int i, j ,k, l, m, bonds, max_s2;
  int run_option;
  
  char root_name[100],append[]=".dat";
  char path[]="../Data/";
  char tempPath[100];
  // char path[100];
  char input_filename[100];
  char temp[100];
  char configurations2[100];

  int dummyInt,dummyInt2;
  float dummyFloat;

  char dummy;

  int typeArrayMax[25]={0,5,2,9,2,9,3,3,1,19,1,3,4,1,0,0};
  int **typeArray;


  typeArrayMax[1]=5;typeArrayMax[2]=2;typeArrayMax[3]=9;
  typeArrayMax[4]=2;typeArrayMax[5]=9;typeArrayMax[6]=3;
  typeArrayMax[7]=3;typeArrayMax[8]=1;typeArrayMax[9]=19;

  typeArrayMax[10]=1;typeArrayMax[11]=3;typeArrayMax[12]=4;
  typeArrayMax[13]=1;typeArrayMax[14]=0;typeArrayMax[15]=0;

  typeArray=imatrix(0,25,0,25);

  setArray(typeArray);

  Control.file=1;    // tells program to save CCM equations to file.

  if (argc == 3) {

    strcpy(input_filename,path);
    strcat(input_filename,argv[1]);
    run_option=atoi(argv[2]);
    if (myid==0)
      printf("myid=%d name=%s runoption=%d\n",myid,input_filename,run_option);

  } else {
    
    printf("Input the 1) root CCM path and 2) name of the CCM script file!"
	   "\nExiting!!\n\n");
    exit(0);
    
  }
  
  if (myid==0)
  printf("/-------------------------------------------------\\\n"
	 "|                                                 |\n"
	 "|                                                 |\n"
	 "|     /=====\\     /=====\\     |\\          /|      |\n"
	 "|    /           /            | \\        / |      |\n"
	 "|    |           |            |  \\      /  |      |\n"
	 "|    |           |            |   \\    /   |      |\n"
	 "|    \\           \\            |    \\  /    |      |\n"
	 "|     \\=====/     \\=====/     |     \\/     |      |\n"
	 "|                                                 |\n"
	 "|                                                 |\n"
	 "|           =====>>>>> 2002 <<<<=====             |\n"
	 "|                                                 |\n"
	 "\\-------------------------------------------------/\n"
	 "\n"
	 "===================================================\n"
	 "Name:       ccm.C\n"
	 "@Copyright: Damian J. J. Farnell\n"
	 "Date:       Wed Jul 17 12:03:05 BST 2002\n"
	 "Purpose:    Determine and solve CCM LSUBm equations.\n"
	 "===================================================\n\n");

  script=fopen(input_filename,"r");
  if (!script) {
    printf("Cannot Open %20s ...\nExiting!!\n\n",input_filename);
    exit(1);
  }

  // read in `control' parameters first:

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "Control Parameters:\n");

  moveAlongHome(); Control.output=readInInteger();   // output level= ...

  if (Control.output > 0 && myid ==0)
    printf("Output (0-non to >6 verbose) =%d\n",Control.output);
 
  moveAlongHome(); fscanf(script,"%80s",tempPath);  // work directory= ...

  moveAlongHome(); fscanf(script,"%80s",temp);      // output file= ...
  
  memcpy(Control.name,tempPath,strlen(tempPath));
#if (PARALLEL_FLAG>0)
  char processorLabel[100]="";
  convert_numeral_to_char(myid,processorLabel);
  strcat(Control.name,"myid_");
  strcat(Control.name,processorLabel);
  strcat(Control.name,"_");
#endif
  strcat(Control.name,temp);

  if (myid==0)
  printf("Simulation Filename = %30s : myid == %d\n",Control.name,myid);

  // read in parameters for solve.C

  
  // now read in the `solve' variables

  moveAlongHome();
  minimum=readInFloat();
  if (Control.output > 0 && myid ==0)
    printf("Minimum delta=%20.10f\n",minimum);

  moveAlongHome();
  maximum=readInFloat();
  if (Control.output > 0 && myid ==0)
    printf("Maximum delta=%20.10f\n",maximum);

  moveAlongHome();
  increment=readInFloat();
  if (Control.output > 0 && myid ==0)
    printf("Increment delta=%30.20f\n",increment);

  // next two variables are minimum and maximum y values
  // used in plot.tcl but not in ccm C program.

  moveAlongHome();
  dummyFloat=readInFloat();
  moveAlongHome();
  dummyFloat=readInFloat();

  // read in highest level parameters first:

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	 "General Parameters:\n");
  
  // dimension

  moveAlongHome();
  Params.dim=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("Dimension=%d\n",Params.dim);

  // spin number

  moveAlongHome();
  dummyFloat = readInFloat();
  // Joerg: We store the twofold of Spin quantum number to have an integer
  //        this is faster and makes the program more compact
  // ToDo: its the maximum of spins in the unit cell and can be
  //       automatically calculated (see below), so its obsolete to set here
  Params.s2 = (int) (2*dummyFloat + 0.5);
  if (Control.output > 0 && myid ==0)
    printf("Maximum spin quantum number=%f\n",Params.s2 / 2.0);

  initDivisors();  // store 0! .. (2s+1)! into array

  // Bounding Triangle

  dummyInt=0;

  moveAlongHome();
  Params.btLength=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("BT Length=%d\n",Params.btLength);  // Joerg: ToDo: whats means BT?
  
  // SUBn-m: n=

  moveAlongHome();
  Params.n=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("SUBn-m: n=%d\n",Params.n);  // flips ???
  
  // SUBn-m: m=

  moveAlongHome();
  Params.m=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("SUBn-m: m=%d\n",Params.m);  // maximum distance ???
  
  // Number of Bonds: 

  moveAlongHome();
  Params.bonds=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("Bonds: m=%d\n",Params.bonds);

  if (Params.bonds > MAX_BONDS) { // defined in Externals.h
    printf("Error input: Bonds > MAX_BONDS\n");
    exit(1);
  }

  /* array of parameters depending on one variable Parameter
   *  sample: d1=cos(d), d2=sin(d) with H= d1*H1+d2*H2
   */ 
  delta = (double *) malloc((Params.bonds+1)*sizeof(double));
  if (!delta) {
    printf("input.C: delta = malloc() failed...\nExit!\n");
    exit(1);
  }
  
  moveAlongHome();
  restrictionClusters=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("Restriction on Clusters: Type=%d\n",restrictionClusters);
  
  moveAlongHome();
  solutionStrategy=readInInteger();
  if (Control.output > 0 && myid ==0)
    if (solutionStrategy == 1)
      printf("Solve CCM equations using Newton-Raphson method\n");
    else
      printf("Solve CCM equations using Direct Iteration method\n");
  
  if (PARALLEL_FLAG > 0 && solutionStrategy == 1) {
    if (myid==0)
      printf("PARALLEL_FLAG > 0 && solutionStrategy == 1...\n"
	     "Exiting!!\n\n");
    exit(1);
  }

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Unit Cell:\n");
  
  // number of different elements of a symmetric multidimensional matrix
  if (Params.dim == 1) 
    max_number_cells_BT=Params.btLength;
  else if (Params.dim == 2)
    max_number_cells_BT=Params.btLength*(Params.btLength+1)/2;
  else {
    // Joerg Sep05, ToDo: use dimension independent formular
    max_number_cells_BT=Params.btLength*(Params.btLength+1)/2
                                       *(Params.btLength+2)/3;
  }

  if (Control.output > 1&& myid ==0) 
    printf("Dimension=%d\ns=%5.3f\nm=%d\nB.T.Length=%d\n"
	   "#Sites in B.T.=%d\n\n",
	   Params.dim,
	   Params.s2 / 2.0,
	   Params.m,
	   Params.btLength,
	   max_number_cells_BT);

  // 1 <= max_number_cells_BT <= MAX_BT  (Joerg Sep05)

  if ( max_number_cells_BT > MAX_BT) {
    printf("Number of Cells in BT is greater than MAX_BT...\n"
	  "Exiting!\n\n");
    exit(1);
  }

  // probably a good idea to keep 
  // Params.btLength >= Params.m/UnitCell.number+1;

  for(i=0;i<MAX_UNIT_CELL;i++) {
    UnitCell.x[i]=0.0;
    UnitCell.y[i]=0.0;
    UnitCell.z[i]=0.0;
    UnitCell.s2[i]=1; // 2*s
  }

  moveAlongHome();
  fscanf(script,"%d",&UnitCell.number);
  if (Control.output > 0 && myid ==0)
    printf("Number of sites in Unit Cell = %d\n",UnitCell.number);
  if (Control.output > 0 && myid ==0)
    printf("Positions & Spin Quantum Number:\n");

  // read in all sites of the unit cell
  for(max_s2=1,i=1;i<=UnitCell.number;i++) {
    
    UnitCell.x[i]=readInFloat();
    UnitCell.y[i]=readInFloat();
    UnitCell.z[i]=readInFloat();
    UnitCell.s2[i]=(int)(2*fabs(readInFloat())+0.5);
    //    moveAlongHome();
    if (max_s2 < UnitCell.s2[i])
        max_s2 = UnitCell.s2[i];

  }

  // check that spin length doesnt exceed maxspinlength (Joerg 2005)
  if (max_s2 != Params.s2) {
     printf("WARNING input: (max_spin=%3.1f) !=  (spin=%3.1f),"
            " set spin automatically to max_spin\n",
              0.5*max_s2, 0.5*Params.s2);
     Params.s2 = max_s2;
     initDivisors();  // store 0! .. (2s+1)! into array
     // exit(1); // We dont need to worry, Joerg
  }

  for(i=1;i<=UnitCell.number;i++) {
    if (Control.output > 0 && myid ==0)
      printf("Site=%d : position = (%f,%f,%f) : spin = %f\n",i,
	     UnitCell.x[i],
	     UnitCell.y[i],
	     UnitCell.z[i],
	     0.5*UnitCell.s2[i]);
  }

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Bravais Lattice:\n");

  // define the bravais lattice 

  moveAlongHome();
  BravaisLattice.a[1]=readInFloat();
  BravaisLattice.a[2]=readInFloat();
  BravaisLattice.a[3]=readInFloat();
  BravaisLattice.b[1]=readInFloat();
  BravaisLattice.b[2]=readInFloat();
  BravaisLattice.b[3]=readInFloat();
  BravaisLattice.c[1]=readInFloat();
  BravaisLattice.c[2]=readInFloat();
  BravaisLattice.c[3]=readInFloat();
  
  if (Control.output > 0 && myid ==0) {
    printf("Bravais Lattice Vector A=(%6.4f,%6.4f,%6.4f)\n",
	   BravaisLattice.a[1],BravaisLattice.a[2],BravaisLattice.a[3]);
    printf("Bravais Lattice Vector B=(%6.4f,%6.4f,%6.4f)\n",
	   BravaisLattice.b[1],BravaisLattice.b[2],BravaisLattice.b[3]);
    printf("Bravais Lattice Vector C=(%6.4f,%6.4f,%6.4f)\n",
	   BravaisLattice.c[1],BravaisLattice.c[2],BravaisLattice.c[3]);
  }

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "Nearest-Neighbours to sites in Unit Cell.:\n");
  

  for(i=1;i<=UnitCell.number;i++) {
    moveAlongHome();
    LatticeNeighbours.number[i]=readInInteger(); 
    for(j=1;j<=LatticeNeighbours.number[i];j++) {
	  LatticeNeighbours.x[i][j]=readInInteger();
	  LatticeNeighbours.y[i][j]=readInInteger();
	  LatticeNeighbours.z[i][j]=readInInteger();
	  LatticeNeighbours.cell[i][j]=readInInteger();
    }
  }

  if (Control.output > 0 && myid ==0) {
    for(i=1;i<=UnitCell.number;i++) {
      printf("Number of nearest-neighbours for site =%d : \n",
	     LatticeNeighbours.number[i]);
      for(j=1;j<=LatticeNeighbours.number[i];j++) {
	printf("#%d : Bravais translation=(%d,%d,%d) to cell (%d)\n",
	       j, LatticeNeighbours.x[i][j],LatticeNeighbours.y[i][j],
	       LatticeNeighbours.z[i][j],LatticeNeighbours.cell[i][j]);
      }
    }
    printf("=======================================\n"
	   "The symmetry operations:\n");
  }
  moveAlongHome();

  LatticeSymmetries.number=readInInteger(); 

  // just to make sure set to sero initially.

  for(i=1;i<=LatticeSymmetries.number;i++){
    LatticeSymmetries.x[i]=0.0;
    LatticeSymmetries.y[i]=0.0;
    LatticeSymmetries.z[i]=0.0;
    for(j=1;j<=3;j++){
      for(k=1;k<=3;k++){
	LatticeSymmetries.R[i][j][k]=0.0;
      }
    }
  }

  for(i=1;i<=LatticeSymmetries.number;i++) {
    for(j=1;j<=3;j++){
      for(k=1;k<=3;k++){
	LatticeSymmetries.R[i][j][k]=readInFloat();
      }
      if (j==1)	
	LatticeSymmetries.x[i]=readInFloat();
      else if (j==2) 
	LatticeSymmetries.y[i]=readInFloat();
      else 
	LatticeSymmetries.z[i]=readInFloat();
    }
  }

  if (Control.output > 0 && myid ==0) {
    for(i=1;i<=LatticeSymmetries.number;i++) {
      printf("Symmetry Operation = %d\nRotation matrix : Prim. Transl. \n",i);
      for(j=1;j<=3;j++){
	for(k=1;k<=3;k++){
	  printf("% 5.3f ", LatticeSymmetries.R[i][j][k]);
	}
	if (j==1)	
	  printf(" : %5.3f\n",LatticeSymmetries.x[i]);
	else if (j==2) 
	  printf(" : %5.3f\n",LatticeSymmetries.y[i]);
	else 
	  printf(" : %5.3f\n",LatticeSymmetries.z[i]);
      }
    }
  }

  // now determine the Hamiltonian. The `types' of term
  // first...
  
  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Hamiltonian:\n");
  
  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(i=0;i<MAX_UNIT_CELL;i++) {
      for(j=0;j<MAX_NUM_NEIGHBOURS;j++) {
	for(k=0;k<15;k++) {  // array defined as [200], why? Joerg05
	  Hamiltonian_types[bonds][k].k_to_m_number[i]=0;
	  Hamiltonian_types[bonds][k].m_to_k_number[i]=0;
	  Hamiltonian_types[bonds][k].k_to_m[i][j][0]=0; // x
	  Hamiltonian_types[bonds][k].k_to_m[i][j][1]=0; // y
	  Hamiltonian_types[bonds][k].k_to_m[i][j][2]=0; // z
	  Hamiltonian_types[bonds][k].k_to_m[i][j][3]=0; // cell
	  Hamiltonian_types[bonds][k].m_to_k[i][j][0]=0; // x
	  Hamiltonian_types[bonds][k].m_to_k[i][j][1]=0; // y
	  Hamiltonian_types[bonds][k].m_to_k[i][j][2]=0; // z
	  Hamiltonian_types[bonds][k].m_to_k[i][j][3]=0; // cell
	  Hamiltonian_types[bonds][k].coefficient[i][j]=0.0;
	}
      }
    }
  }

  moveAlongHome();
  directional=readInInteger();

  if (myid==0)
    printf("Directional?=%d\n",directional);

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(i=1;i<=UnitCell.number;i++) {
      for(k=1;k<14;k++) {    /* 1..13 */

	moveAlongHome();
	Hamiltonian_types[bonds][k].k_to_m_number[i]=readInInteger();
	
	if (k<=9) {          /* 1..9 */

	  if (directional) {
	    for(j=1;j<=Hamiltonian_types[bonds][k].k_to_m_number[i];j++) {
	      Hamiltonian_types[bonds][k].k_to_m[i][j][0]=readInInteger();
	      Hamiltonian_types[bonds][k].k_to_m[i][j][1]=readInInteger();
	      Hamiltonian_types[bonds][k].k_to_m[i][j][2]=readInInteger();
	      Hamiltonian_types[bonds][k].k_to_m[i][j][3]=readInInteger();
	      Hamiltonian_types[bonds][k].coefficient[i][j]=readInFloat();
	    }

	    moveAlongHome();
	    Hamiltonian_types[bonds][k].m_to_k_number[i]=readInInteger();

	    for(j=1;j<=Hamiltonian_types[bonds][k].m_to_k_number[i];j++) {
	      Hamiltonian_types[bonds][k].m_to_k[i][j][0]=readInInteger(); // x
	      Hamiltonian_types[bonds][k].m_to_k[i][j][1]=readInInteger(); // y
	      Hamiltonian_types[bonds][k].m_to_k[i][j][2]=readInInteger(); // z
	      Hamiltonian_types[bonds][k].m_to_k[i][j][3]=readInInteger(); // cell
	      Hamiltonian_types[bonds][k].coefficient[i][j]=readInFloat();
	    }

	  } else {    /* 10..13 */

	    Hamiltonian_types[bonds][k].m_to_k_number[i]=
	      Hamiltonian_types[bonds][k].k_to_m_number[i];

	    for(j=1;j<=Hamiltonian_types[bonds][k].k_to_m_number[i];j++) {

	      Hamiltonian_types[bonds][k].k_to_m[i][j][0]=readInInteger();
	      Hamiltonian_types[bonds][k].k_to_m[i][j][1]=readInInteger();
	      Hamiltonian_types[bonds][k].k_to_m[i][j][2]=readInInteger();
	      Hamiltonian_types[bonds][k].k_to_m[i][j][3]=readInInteger();
	      Hamiltonian_types[bonds][k].coefficient[i][j]=readInFloat();

	      Hamiltonian_types[bonds][k].m_to_k[i][j][0]=Hamiltonian_types[bonds][k].k_to_m[i][j][0];
	      Hamiltonian_types[bonds][k].m_to_k[i][j][1]=Hamiltonian_types[bonds][k].k_to_m[i][j][1];
	      Hamiltonian_types[bonds][k].m_to_k[i][j][2]=Hamiltonian_types[bonds][k].k_to_m[i][j][2];
	      Hamiltonian_types[bonds][k].m_to_k[i][j][3]=Hamiltonian_types[bonds][k].k_to_m[i][j][3];	      

	    }
	  }
	} else {
	  Hamiltonian_types[bonds][k].coefficient[i][1]=readInFloat();
	}
      }
    }
  }

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(k=1;k<14;k++) {
      Hamiltonian_types[bonds][k].evaluate=0;
      for(i=1;i<=UnitCell.number;i++) {
	for(j=1;j<=Hamiltonian_types[bonds][k].k_to_m_number[i];j++) {
	  if (fabs(Hamiltonian_types[bonds][k].coefficient[i][j]) > 0.01) {
	    Hamiltonian_types[bonds][k].evaluate=1;
	    break;
	  }
	}
	if (Hamiltonian_types[bonds][k].evaluate) break;
      }
    }
  }

  // now print out this information:
    
  if (Control.output > 0 && myid ==0) {
    for(bonds=1;bonds<=Params.bonds;bonds++) {
      for(k=1;k<14;k++) {
	if (Control.output>0) 
	  printf("Bond = %d : Type = %d : Evaluate =%d\n"
		 ,bonds,k, Hamiltonian_types[bonds][k].evaluate);
	if (Control.output>3) {
	  if (k > 9) {
	    for(i=1;i<=UnitCell.number;i++) {
	      printf("Site = %d : # Single Site Term : ",i);
	      printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		     Hamiltonian_types[bonds][k].k_to_m[i][1][0],
		     Hamiltonian_types[bonds][k].k_to_m[i][1][1],
		     Hamiltonian_types[bonds][k].k_to_m[i][1][2],
		     Hamiltonian_types[bonds][k].k_to_m[i][1][3],
		     Hamiltonian_types[bonds][k].coefficient[i][1]);
	    }
	  } else {

	    for(i=1;i<=UnitCell.number;i++) {

	      if (directional) {
		
		printf("Site = %d : Number Directional Neighbours k to m = %d\n",
		       i,Hamiltonian_types[bonds][k].k_to_m_number[i]);
		
		for(j=1;j<=Hamiltonian_types[bonds][k].k_to_m_number[i];j++) {
		  printf("(%d,%d,%d,%d) : Coeff. =%f\n",
			 Hamiltonian_types[bonds][k].k_to_m[i][j][0],
			 Hamiltonian_types[bonds][k].k_to_m[i][j][1],
			 Hamiltonian_types[bonds][k].k_to_m[i][j][2],
			 Hamiltonian_types[bonds][k].k_to_m[i][j][3],
			 Hamiltonian_types[bonds][k].coefficient[i][j]);
		}
		printf("Site = %d : Number Directional Neighbours m to k = %d\n",i,
		       Hamiltonian_types[bonds][k].m_to_k_number[i]);
		for(j=1;j<=Hamiltonian_types[bonds][k].m_to_k_number[i];j++) {
		  printf("(%d,%d,%d,%d) : Coeff. =%f\n",
			 Hamiltonian_types[bonds][k].m_to_k[i][j][0], // x
			 Hamiltonian_types[bonds][k].m_to_k[i][j][1], // y
			 Hamiltonian_types[bonds][k].m_to_k[i][j][2], // z
			 Hamiltonian_types[bonds][k].m_to_k[i][j][3], // cell
			 Hamiltonian_types[bonds][k].coefficient[i][j]);
		}
	      } else {
		
		printf("Site = %d : Number Non-Directional Neighbours  = %d\n"
		       ,i,Hamiltonian_types[bonds][k].k_to_m_number[i]);
		
		for(j=1;j<=Hamiltonian_types[bonds][k].k_to_m_number[i];j++) {
		  printf("(%d,%d,%d,%d) : Coeff. =%f\n",
			 Hamiltonian_types[bonds][k].k_to_m[i][j][0],
			 Hamiltonian_types[bonds][k].k_to_m[i][j][1],
			 Hamiltonian_types[bonds][k].k_to_m[i][j][2],
			 Hamiltonian_types[bonds][k].k_to_m[i][j][3],
			 Hamiltonian_types[bonds][k].coefficient[i][j]);
		}
	      }
	    }
	  }
	}
      }
    }
  }

  // now input the information for the Hamiltonian proper for each term
  // in the similarity transformed version of H.

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(k=1;k<14;k++) {

      if (Hamiltonian_types[bonds][k].evaluate) {

	for(l=1;l<=typeArrayMax[k];l++) {

          int term = typeArray[k][l];
	  Hamiltonian[bonds][term].evaluate=1;

	  if (k > 9) {
	    for(i=1;i<=UnitCell.number;i++) {
	      Hamiltonian[bonds][term].k_to_m[i][1][0]=Hamiltonian_types[bonds][k].k_to_m[i][1][0];
	      Hamiltonian[bonds][term].k_to_m[i][1][1]=Hamiltonian_types[bonds][k].k_to_m[i][1][1];
	      Hamiltonian[bonds][term].k_to_m[i][1][2]=Hamiltonian_types[bonds][k].k_to_m[i][1][2];
	      Hamiltonian[bonds][term].k_to_m[i][1][3]=Hamiltonian_types[bonds][k].k_to_m[i][1][3];
	      Hamiltonian[bonds][term].coefficient[i][1]=Hamiltonian_types[bonds][k].coefficient[i][1];
	    }
	  } else {

	    for(i=1;i<=UnitCell.number;i++) {
	      
	      Hamiltonian[bonds][term].k_to_m_number[i]=
		Hamiltonian_types[bonds][k].k_to_m_number[i];
	      
	      for(j=1;j<=Hamiltonian_types[bonds][k].k_to_m_number[i];j++) {

		Hamiltonian[bonds][term].k_to_m[i][j][0]=Hamiltonian_types[bonds][k].k_to_m[i][j][0];
		Hamiltonian[bonds][term].k_to_m[i][j][1]=Hamiltonian_types[bonds][k].k_to_m[i][j][1];
		Hamiltonian[bonds][term].k_to_m[i][j][2]=Hamiltonian_types[bonds][k].k_to_m[i][j][2];
		Hamiltonian[bonds][term].k_to_m[i][j][3]=Hamiltonian_types[bonds][k].k_to_m[i][j][3];
		Hamiltonian[bonds][term].coefficient[i][j]=Hamiltonian_types[bonds][k].coefficient[i][j];

	      }
	      
	      Hamiltonian[bonds][term].m_to_k_number[i]=
		Hamiltonian_types[bonds][k].m_to_k_number[i];
	      
	      for(j=1;j<=Hamiltonian_types[bonds][k].m_to_k_number[i];j++) {

		Hamiltonian[bonds][term].m_to_k[i][j][0]=Hamiltonian_types[bonds][k].m_to_k[i][j][0];
		Hamiltonian[bonds][term].m_to_k[i][j][1]=Hamiltonian_types[bonds][k].m_to_k[i][j][1];
		Hamiltonian[bonds][term].m_to_k[i][j][2]=Hamiltonian_types[bonds][k].m_to_k[i][j][2];
		Hamiltonian[bonds][term].m_to_k[i][j][3]=Hamiltonian_types[bonds][k].m_to_k[i][j][3];
		Hamiltonian[bonds][term].coefficient[i][j]=Hamiltonian_types[bonds][k].coefficient[i][j];

	      }
	    }
	  }
	}
      }
    }
  }

  if (Control.output>3 && myid ==0) {

    for(bonds=1;bonds<=Params.bonds;bonds++) {
      printf("\n");

      for(k=1;k<63;k++) {

	printf("Bond = %d : Type = %d : Evaluate =%d\n",bonds,k,
	       Hamiltonian[bonds][k].evaluate);      

	if (k > 53) {

	  for(i=1;i<=UnitCell.number;i++) {
	    printf("Site = %d : # Single Site Term : ",i);
	    printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		   Hamiltonian[bonds][k].k_to_m[i][1][0],
		   Hamiltonian[bonds][k].k_to_m[i][1][1],
		   Hamiltonian[bonds][k].k_to_m[i][1][2],
		   Hamiltonian[bonds][k].k_to_m[i][1][3],
		   Hamiltonian[bonds][k].coefficient[i][1]);
	  }

	} else {

	  for(i=1;i<=UnitCell.number;i++) {

	    if (directional) {

	      printf("Site = %d : Number Directional Neighbours k to m = %d\n",
		     i,Hamiltonian[bonds][k].k_to_m_number[i]);
	      for(j=1;j<=Hamiltonian[bonds][k].k_to_m_number[i];j++) {
		printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		       Hamiltonian[bonds][k].k_to_m[i][j][0],
		       Hamiltonian[bonds][k].k_to_m[i][j][1],
		       Hamiltonian[bonds][k].k_to_m[i][j][2],
		       Hamiltonian[bonds][k].k_to_m[i][j][3],
		       Hamiltonian[bonds][k].coefficient[i][j]);
	      }
	      printf("Site = %d : Number Directional Neighbours m to k = %d\n",i,
		     Hamiltonian[bonds][k].m_to_k_number[i]);
	      for(j=1;j<=Hamiltonian[bonds][k].m_to_k_number[i];j++) {
		printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		       Hamiltonian[bonds][k].m_to_k[i][j][0],
		       Hamiltonian[bonds][k].m_to_k[i][j][1],
		       Hamiltonian[bonds][k].m_to_k[i][j][2],
		       Hamiltonian[bonds][k].m_to_k[i][j][3],
		       Hamiltonian[bonds][k].coefficient[i][j]);
	      }

	    } else {

	      printf("Site = %d : Number Non-Directional Neighbours  = %d\n"
		     ,i,Hamiltonian[bonds][k].k_to_m_number[i]);

	      for(j=1;j<=Hamiltonian[bonds][k].k_to_m_number[i];j++) {
		printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		       Hamiltonian[bonds][k].k_to_m[i][j][0],
		       Hamiltonian[bonds][k].k_to_m[i][j][1],
		       Hamiltonian[bonds][k].k_to_m[i][j][2],
		       Hamiltonian[bonds][k].k_to_m[i][j][3],
		       Hamiltonian[bonds][k].coefficient[i][j]);
	      }

	      printf("Site = %d : Number Non-Directional Neighbours  = %d\n"
		     ,i,Hamiltonian[bonds][k].m_to_k_number[i]);

	      for(j=1;j<=Hamiltonian[bonds][k].k_to_m_number[i];j++) {
		printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		       Hamiltonian[bonds][k].m_to_k[i][j][0],
		       Hamiltonian[bonds][k].m_to_k[i][j][1],
		       Hamiltonian[bonds][k].m_to_k[i][j][2],
		       Hamiltonian[bonds][k].m_to_k[i][j][3],
		       Hamiltonian[bonds][k].coefficient[i][j]);
	      }
	    }
	  }
	}
      }
    }
  }

  fclose(script);
  free_imatrix(typeArray,0,25,0,25);
  
  definePointsInBT();

  //  init x_array[],x_index[],cell_index[]
  //  defineGeneralisedPositions();  -> removed (not used, Joerg05)

  // not used at moment...

  sublatticeMagnetisation.number=UnitCell.number;
  for (i=1;i<=sublatticeMagnetisation.number;i++)
    sublatticeMagnetisation.cell[j]=1;
 
  // read in cluster configurations 
  
  if (run_option == 2 || run_option == 3) {

    strcpy(configurations2,Control.name);
    strcat(configurations2,".configs");
    fp1=fopen(configurations2,"r");  
    if (!fp1) {
      printf("Cannot Open %s...\nExiting!!\n\n",configurations2);
      exit(1);
    }

    fread(&NumberConfigurations,sizeof(int),1,fp1);
    
    if (Control.output > 1) {
      printf("\n==========================================================\n"
	     "Fundamental clusters (N_F = %d) are: \n"
	     "============================================================\n\n"
	     ,NumberConfigurations);
    }

    for(i=1;i<=NumberConfigurations;i++) {

      if (Control.output > 1 && myid==0)
	printf("Cluster %d: \n",i);

      fread(spinConfigurations[i].descriptor,
	    ((MAX_SPIN_LIST)*sizeof(int)),
	    1,
	    fp1);
      if (Control.output > 1 && myid==0){
	printOutClusterFromDescriptor(spinConfigurations[i].descriptor);
	printf("\n");
      }
    }
    fclose(fp1);

    /* NEW!!: work out the maximum radius for all clusters 
       This is used to reduce the amount of sweeping for 
       F_k F_m s_k^z s_m^z -like terms. */

    radius=0;
   
    {
       int length,descriptorIndex;
      
      int **tempConfig;
      tempConfig=imatrix(0,MAX_LSUBM+1,0,5);
      
      for(i=1;i<=NumberConfigurations;i++) {
	
	length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
					 tempConfig);
	
	/* note that midPointX, midPointY, midPointZ indicate Unit cells */
	
	// note that integer rounding shouldn't affect this 
	// algorithm
	
	// now find the maximum `radius'
	
	for(j=1;j<=length;j++) {
	  if (abs(tempConfig[j][1]) > radius) radius=abs(tempConfig[j][1]);
	  if (abs(tempConfig[j][2]) > radius) radius=abs(tempConfig[j][2]);
	  if (abs(tempConfig[j][3]) > radius) radius=abs(tempConfig[j][3]);
	}
      }
      
      radius+=1;
      
      printf("\nRadius = %d\n\n",radius);
     
      free_imatrix(tempConfig,0,MAX_LSUBM+1,0,5);
 
    }
  }
}


void moveAlongHome(void)
{
  char dummy, buff[400];
  int  count, bp=0, iscomment=0;

  /* 
     purpose of this routine is to scan along the file until
     an `=' sign is seen 
     Joerg Sep05: ignore out commented lines (// ... and # ...)
  */
  
  for(count=0;count<400;count++) {
    fscanf(script,"%c",&dummy);
    buff[bp]=dummy; bp++; buff[bp]=0;
    if (dummy == '\n' || dummy == '=') {
       if (bp) buff[bp-1]='\n';
       if (Control.output > 0 && myid == 0) printf("%% %s",buff);
       bp=0; buff[bp]=0;
    }
    if (bp>1 && buff[bp-1] == '/' && buff[bp-2]== '/') iscomment=1;
    if (dummy == '#')  iscomment=1;
    if (dummy == '\n') iscomment=0;
    if (dummy == '=' && !iscomment) break;
  }

  if (count >=400) {
    printf("Error in reading CCM script: can't find delimiter in time!\n"
	   "Exiting!!\n\n");
    exit(1);
  }

}

int readInInteger(void)
{
  int sum;
  fscanf(script,"%d",&sum);
  return sum;
}
    
double readInFloat(void)
{
  double sum;
  fscanf(script,"%lf",&sum);
  return sum;
};


// ToDo: what ist it for?  Joerg Sep05
void setArray(int **typeArray)
{

  int i,j;
    
  for(i=1;i<=13;i++)
    for(j=1;j<=19;j++)
      typeArray[i][j]=0;

  typeArray[1][1]=1; typeArray[1][2]=2; typeArray[1][3]=3;typeArray[1][4]=4;
  typeArray[1][5]=5;

  typeArray[2][1]=6; typeArray[2][2]=7;

  typeArray[3][1]=8; typeArray[3][2]=9; typeArray[3][3]=10;typeArray[3][4]=11;
  typeArray[3][5]=12;typeArray[3][6]=13;typeArray[3][7]=46;
  typeArray[3][8]=47;typeArray[3][9]=48;

  typeArray[4][1]=14;typeArray[4][2]=15;

  typeArray[5][1]=16;typeArray[5][2]=17;typeArray[5][3]=18;typeArray[5][4]=19;
  typeArray[5][5]=20;typeArray[5][6]=21;typeArray[5][7]=49;typeArray[5][8]=50;
  typeArray[5][9]=51;

  typeArray[6][1]=22;typeArray[6][2]=23;typeArray[6][3]=52;

  typeArray[7][1]=24;typeArray[7][2]=25;typeArray[7][3]=53;
  
  typeArray[8][1]=26;

  for(i=27;i<=45;i++)
    typeArray[9][i-26]=i;
  
  typeArray[10][1]=54;

  typeArray[11][1]=55;typeArray[11][2]=56;typeArray[11][3]=57;

  typeArray[12][1]=58;typeArray[12][2]=59;typeArray[12][3]=60;
  typeArray[12][1]=61;

  typeArray[13][1]=62;
}

// called by input()
// initialize: lookUpTable used by [printOut,find]ClusterFromDescriptor
//             BTarray     used by findConfigurationToDescriptor
// ToDo: estimate size and malloc lookUpTable, simplify (Joerg)
void definePointsInBT(void)
{

  int i,j,k,l,m,count;

  for(m=0;m<Params.btLength;m++) 
    for(l=0;l<Params.btLength;l++) 
      for(k=0;k<Params.btLength;k++) 
	BTarray[k][l][m]=-99;
  
  for(m=0;m<MAX_BT;m++) {  // default: 20000
    lookUpTable[m][1]=0;
    lookUpTable[m][2]=0;
    lookUpTable[m][3]=0;
  }

  if (Params.dim == 1) {
    for(m=0;m<Params.btLength;m++) {
      lookUpTable[m][1]=m;    // = count+1 (but unused)
      lookUpTable[m][2]=0;
      lookUpTable[m][3]=0;
    }
  } else if (Params.dim == 2) {
    count=0;
    for (l=0;l<Params.btLength  ;l++)
    for (k=0;k<Params.btLength-l;k++) {  // only upper left half matrix
        // sample: btLength=4
        // (l,k)=(0,0),(0,1),(0,2),(0,3),  [x]
        //       (1,0),(1,1),(1,2),
        //       (2,0),(2,1),
        //       (3,0)                   n*(n+1)/2 = 4*5/2 = 10 entries
        //       [y]
	count++;
	BTarray[k][l][0]=count; 
	lookUpTable[count][1]=k;
	lookUpTable[count][2]=l;
	lookUpTable[count][3]=0;
    }
  } else if (Params.dim == 3) {
    count=0;
    for (m=0;m<Params.btLength    ;m++) 
    for (l=0;l<Params.btLength-m  ;l++) 
    for (k=0;k<Params.btLength-l-m;k++) { // replacable by a function ???
	count++;                    //  n*(n+1)*(n+2)/6
	BTarray[m][l][k]=count;     // len=10 -> old=385 new=220
	lookUpTable[count][1]=m;
	lookUpTable[count][2]=l;
	lookUpTable[count][3]=k;
    }
  }
  // added by Joerg Sep2005  cubic 10 -> 10*12^2/4=1440/4=385=10*11/2*7
  if (Control.output > 0)
  printf("definePointsInBT: length= %d sites= %d\n", Params.btLength, count);
}  

